import warnings
import logging
import openpyxl
from openpyxl.utils import get_column_letter
from openpyxl.styles import NamedStyle, Border, Side, Alignment, PatternFill
from openpyxl.formatting import Rule
from openpyxl.styles.differential import DifferentialStyle

from . import lex


class LExcel:

    CODE_ROW = 6      # Row of board code
    CODE_COL = 1      # Column of the first board code
    MOVES_ROW = 8     # Row of the first move
    MOVES_OFFSET = 0  # Offset of moves w.r.t. CODE_COL

    STYLE = NamedStyle(name="default")
    BORDER = Side(style='thin', color="000000")
    STYLE.border = Border(left=BORDER, top=BORDER, right=BORDER, bottom=BORDER)
    STYLE.alignment = Alignment(horizontal="center", vertical="center")
    CYAN_STYLE = Rule(type='cellIs', operator='equal',
                      dxf=DifferentialStyle(
                          fill=PatternFill(bgColor='00FFFF')),
                      formula=[f"{lex.CYAN}"])
    VIOLET_STYLE = Rule(type='cellIs', operator='equal',
                        dxf=DifferentialStyle(
                            fill=PatternFill(bgColor='EE82EE')),
                        formula=["{lex.VIOLET}"])

    def __init__(self, filename: str):
        self.FILE = filename
        warnings.simplefilter("ignore")
        self.WB = openpyxl.load_workbook(self.FILE,
                                         read_only=True,
                                         data_only=True)

    def find_board_column(self, turn: lex.Player, code: str) -> int | None:
        """Return the column of the board identified by code and turn.

        None if code is not found.
        """
        ws = self.WB[f'Turn{turn}']
        ws.calculate_dimension()  # Google sheets save as unsized
        for col in range(self.CODE_COL, ws.max_column+1):
            c = ws.cell(row=self.CODE_ROW, column=col)
            if c.value is not None:
                if c.value == code:
                    return c.column
        return None

    def get_moves(self, board: lex.Board,
                  turn: lex.Player,
                  player: lex.Player = lex.VIOLET) \
            -> tuple[list[str], list[int]] | None:
        """Get the moves associated to a given board and turn.

        Return a pair of lists: the possible moves and the current weights,
        None if the board is not present in filename.
        Illegal moves are always implicitly weighted 0,
        thus they are never returned, rewarded or punished.
        """
        col = self.find_board_column(turn, board.get_code())
        if col is not None:
            moves_col = col + self.MOVES_OFFSET
            moves, weights = [], []
            ws = self.WB['Turn{}'.format(turn)]
            for r in range(self.MOVES_ROW,
                           self.MOVES_ROW+len(board.WELL_FORMED_MOVES)+1):
                move_cell = ws[get_column_letter(moves_col)+str(r)]
                weight_cell = ws[get_column_letter(moves_col+1)+str(r)]
                if move_cell.value is not None:
                    legal_moves = board.get_legal_moves(player)
                    if move_cell.value in legal_moves:
                        moves.append(move_cell.value)
                        assert weight_cell.value is not None, \
                            "None weight %s %s" % (board.get_code(), turn)  # pragma: no mutate
                        weights.append(weight_cell.value)
                    else:
                        moves.append(move_cell.value)
                        weights.append(0)

            return moves, weights
        return None

    def reward(self, newfile: str,
               match_moves: list[tuple[str, str, lex.Player]],
               prize: int):
        """Write a new Excel file with every move in match_moves rewarded.
        """
        warnings.simplefilter("ignore")
        wb = openpyxl.load_workbook(self.FILE, data_only=True)
        for move, code, turn in match_moves:
            ws = wb[f'Turn{turn}']
            col = self.find_board_column(turn, code)
            board = lex.Board(code)
            if col is not None:
                moves_col = col + self.MOVES_OFFSET
                for r in range(self.MOVES_ROW,
                               self.MOVES_ROW+len(board.WELL_FORMED_MOVES)+1):
                    move_cell = ws[get_column_letter(moves_col)+str(r)]
                    weight_cell = ws[get_column_letter(moves_col+1)+str(r)]
                    if move_cell.value == move:
                        weight_cell.value = weight_cell.value + prize
                        break
        wb.save(newfile)
        wb.close()

    def choose_and_remember_move(self, board: lex.Board,
                                 turn: lex.Player,
                                 match_moves: list[tuple[str,
                                                         str,
                                                         lex.Player]],
                                 rand,
                                 player: lex.Player = lex.VIOLET):
        """Randomly choose a move, append info in match_moves to reward a win.

        """

        def choose_move(b: lex.Board):
            moves_weights = self.get_moves(b, turn, player)
            if moves_weights is not None:
                moves, weights = moves_weights
                if sum(weights) > 0:
                    m = rand.choices(moves, weights)
                    logging.debug('Choices: %s %s %s',
                                 moves, weights, m)  # pragma: no mutate
                else:
                    m = moves[0:1]
                return m[0]
            return None

        move = choose_move(board)
        if move is not None:
            match_moves.append((move, board.get_code(), turn))
            return move

        move = choose_move(board.flip())
        if move is not None:
            match_moves.append((move, board.flip().get_code(), turn))
            return board.flip_move(move)

    @classmethod
    def make_empty_experience(cls, ncols: int,
                              filename: str = 'exapawn-empty.xls',
                              players: list[lex.Player] =
                              [lex.VIOLET, lex.CYAN]):
        board = lex.Board(ncols=ncols)
        _, boards = board.get_forest(len(board.COLS))
        wb = openpyxl.Workbook()
        for player in players:
            maxturn = 0
            for t, _ in boards.values():
                if t > maxturn:
                    maxturn = t
            for turn in range(player, maxturn+1, 2):
                name = f'Turn{turn}'
                ws = wb.create_sheet(name)
                code_col = cls.CODE_COL
                for code in sorted(boards):  # sort for reproducibility
                    t, moves = boards[code]
                    if t == turn:
                        board = lex.Board(code)
                        for i, c in enumerate(board.COLS):
                            col = code_col + i
                            x = ws.cell(row=1, column=col, value=c)
                            x.style = cls.STYLE
                            for r in range(2, 2+len(board[c])):
                                x = ws.cell(row=r,
                                            column=col,
                                            value=board[c][r-2])
                                x.style = cls.STYLE
                        x = ws.cell(row=cls.CODE_ROW,
                                    column=code_col,
                                    value=code)
                        x.style = cls.STYLE

                        moves_row = cls.MOVES_ROW
                        for m in sorted(moves):  # sort for reproducibility
                            x = ws.cell(row=moves_row,
                                        column=code_col+cls.MOVES_OFFSET,
                                        value=m)
                            x.style = cls.STYLE
                            x = ws.cell(row=moves_row,
                                        column=code_col+cls.MOVES_OFFSET+1,
                                        value=1)
                            x.style = cls.STYLE
                            moves_row = moves_row + 1

                        code_col = code_col + len(board.COLS) + 1
                ws.conditional_formatting.add(
                    f'A2:{get_column_letter(code_col)}4',
                    cls.CYAN_STYLE)
                ws.conditional_formatting.add(
                    f'A2:{get_column_letter(code_col)}4',
                    cls.VIOLET_STYLE)
        del wb['Sheet']
        wb.save(filename)



from lexcel.game import main, MARKS
import unittest
from unittest.mock import patch
from io import StringIO
import re
import sys
import os
import hashlib


@patch('sys.stdout', new_callable=StringIO)
class TestGame(unittest.TestCase):
    save_in = None

    def setUp(self):
        self.save_in = sys.stdin

    def tearDown(self):
        sys.stdin = self.save_in

    def assert_main_exits(self, args):
        with self.assertRaises(SystemExit) as cm:
            main(args)
        self.assertEqual(cm.exception.code, 0)

    def assert_file_and_rm(self, file):
        self.assertTrue(os.path.isfile(file))
        os.remove(file)

    def test_match(self, mock_out):
        for n in (3,
                  # 4,
                  # 5
                  ):
            for p in (1, 2):
                with self.subTest("Winner {} {}".format(p, n)):
                    sys.stdin = open('tests/winner{}-{}.txt'.format(p, n), 'r')
                    self.assert_main_exits(['-n{}'.format(n)])
                    for player in (1, 2):
                        self.assertIn('\nPlayer {} ({}), make your move: \n'.format(player, MARKS[player]),
                                      mock_out.getvalue())
                    self.assertIn('\nPlayer {} won!!!\n'.format(p), mock_out.getvalue())
                    sys.stdin.close()

    def test_match_exp(self, mock_out):
        sys.stdin = open('tests/7777.txt', 'r')
        self.assert_main_exits(['-e2',
                                'tests/exapawn-empty.xlsx',
                                'exapawn-7777.xlsx',
                                '-b', 'B89', '-p', '1', '-t', '3', '-s', '7777'])
        self.assertIn('Player 2 won', mock_out.getvalue())
        self.assert_file_and_rm('exapawn-7777.xlsx')
        sys.stdin.close()

    def test_empty(self, mock_out):
        for n in (3,
                  # 4,
                  # 5
                  ):
            with self.subTest("{} pawns".format(n)):
                self.assert_main_exits(['-n{}'.format(n),
                                        '--empty',
                                        'exapawn-empty.xlsx'])
                self.assert_file_and_rm('exapawn-empty.xlsx')

    def test_match_empty_exp_seed(self, mock_out):

        def get_digest(filename):
            with open(filename, 'rb') as f:
                hash = hashlib.sha1()
                hash.update(f.read())
            return hash.hexdigest()

        self.assert_main_exits(['--empty', 'exapawn-empty.xlsx'])
        before = get_digest('exapawn-empty.xlsx')

        sys.stdin = open('tests/0001.txt', 'r')
        self.assert_main_exits(['-e2', 'exapawn-empty.xlsx', 'exapawn-test-001.xlsx',
                                '-s', '0001', '-d', 'test0001.log'])
        self.assertIn('Player 2 won', mock_out.getvalue())
        sys.stdin.close()

        after = get_digest('exapawn-empty.xlsx')
        self.assertEqual(before, after)

        sys.stdin = open('tests/0002.txt', 'r')
        self.assert_main_exits(['-e2', 'exapawn-test-001.xlsx', 'exapawn-test-002.xlsx',
                                '-s', '0102'])
        self.assertIn('Player 2 won', mock_out.getvalue())
        sys.stdin.close()

        self.assert_file_and_rm('exapawn-test-001.xlsx')
        self.assert_file_and_rm('exapawn-test-002.xlsx')
        # self.assert_file_and_rm('test0001.log')
        self.assert_file_and_rm('exapawn-empty.xlsx')

    def test_match_empty_exp_seed2(self, mock_out):
        self.assert_main_exits(['--empty', 'exapawn-empty.xlsx'])

        sys.stdin = open('tests/0002c.txt', 'r')
        self.assert_main_exits(['-e2', 'exapawn-empty.xlsx', 'exapawn-test-001c.xlsx',
                                '-s', '0002'])
        self.assertIn('YY move for player 1 is not allowed.\n', mock_out.getvalue())
        self.assertIn('Input is invalid.\n', mock_out.getvalue())
        self.assertIn('\nPlayer 1 won!!!\n', mock_out.getvalue())
        sys.stdin.close()

        self.assert_file_and_rm('exapawn-test-001c.xlsx')
        self.assert_file_and_rm('exapawn-empty.xlsx')

    def test_tree(self, mock_out):
        WIN = re.compile(r'\[.*winner.*\]')
        for n, winners in [
                (3, 65),
                # (4, 1635),
                # (5, 66835)
        ]:
            with self.subTest("{} pawns".format(n)):
                self.assert_main_exits(['-n', str(n), '--tree'])
                count = 0
                for line in mock_out.getvalue().split('\n'):
                    if WIN.search(line) is not None:
                        count = count + 1
                self.assertEqual(count, winners)
                mock_out.truncate(0)


if __name__ == '__main__':
    unittest.main()



"""LEX board operations."""

from functools import cache

type Player = int
type ColName = str
type PawnColumn = dict[ColName, list[Player]]
type Code = str

EMPTY: Player = 0
CYAN: Player = 1
VIOLET: Player = 2

MARKS = {
    EMPTY: ' ',
    CYAN: '\N{WHITE CHESS PAWN}',
    VIOLET: '\N{BLACK CHESS PAWN}'
}


@cache
class Base256:
    """A singleton to implement Base256.
       See https://github.com/fleschutz/Base256U
    """
    SYMBOLS: str
    BASE = 256

    def __init__(self):
        self.SYMBOLS = ''.join([str(i) for i in range(10)])
        self.SYMBOLS += ''.join([chr(ord('A') + i) for i in range(26)])
        self.SYMBOLS += ''.join([chr(ord('a') + i) for i in range(26)])
        start = ord('\N{LATIN CAPITAL LETTER A WITH GRAVE}')
        # Skip non letters
        self.SYMBOLS += ''.join([chr(i)  # 194 symbols in total
                                 for i in range(
                                         start,
                                         start+196)
                                 if chr(i) != '\N{MULTIPLICATION SIGN}'
                                 and chr(i) != '\N{DIVISION SIGN}'])
        assert len(self.SYMBOLS) == self.BASE

    def to_int(self, code: str) -> int:
        assert all([i in self.SYMBOLS for i in code])
        n = 0
        for i, c in enumerate(code[::-1]):
            n += self.BASE**i * self.SYMBOLS.index(c)
        return n

    def from_int(self, n: int) -> str:
        assert n >= 0
        if n == 0:
            return self.SYMBOLS[0]
        s = ''
        x = n
        while x > 0:
            s += self.SYMBOLS[x % self.BASE]
            x = x // self.BASE
        return s


def other(player: Player) -> Player:
    """Return opposite color. """
    if player == VIOLET:
        return CYAN
    if player == CYAN:
        return VIOLET
    return EMPTY


def pawns_to_code(pawns: PawnColumn) -> str:
    """Return an alphanumeric repr. of a board for ease board matching.

    The code is a string of alphanumeric digits representing columns,
    therefore a flipped board has exactly the reverse code.

    """
    r = ''
    base = len(MARKS)
    b256 = Base256()
    for c in sorted(pawns.keys()):
        n = 0
        for i, p in enumerate(pawns[c]):
            n += p * base**i
        r += b256.from_int(n)
    return r


class Board:
    """Boards in a program must have the same number of columns and rows.
    """

    NROWS: int
    COLS: list[ColName] = []
    WELL_FORMED_MOVES: list[str]

    @classmethod
    def _init_board_params(cls, ncols: int = 3):
        """Init globals used for Board instantiation."""
        col_names = ('V', 'W', 'X', 'Y', 'Z')
        assert 0 < ncols <= len(col_names)
        cls.NROWS = ncols
        cls.COLS = list(col_names[-ncols:])
        cls.WELL_FORMED_MOVES = []
        for i in cls.COLS:
            for j in cls.COLS:
                if abs(ord(i) - ord(j)) <= 1:
                    cls.WELL_FORMED_MOVES.append(i+j)

    def _pawns_from_code(self, h: str) -> PawnColumn:
        """Return a dict corresponding to code digits.
        """
        assert len(self.COLS) == len(h), \
            f"{self.COLS} {h}"  # pragma: no mutate
        r = {}
        base = len(MARKS)
        b256 = Base256()
        for i, c in enumerate(self.COLS):
            col = []
            n = b256.to_int(h[i])
            for _ in range(self.NROWS):
                col.append(n % base)
                n = n // base
            r[c] = col

        return r

    def __init__(self,
                 code: str | None = None,
                 ncols: int = 3):
        assert ncols > 0
        base = len(MARKS)
        b256 = Base256()
        if code is None:
            start = b256.from_int(VIOLET + CYAN*base**(ncols-1))
            code = start*ncols
        else:
            ncols = len(code)

        assert ncols == len(code), f'{ncols} {code}'  # pragma: no mutate
        if self.COLS == [] or len(self.COLS) != len(code):
            self._init_board_params(len(code))

        self.pawns = self._pawns_from_code(code)

    def __str__(self) -> str:
        """Return a string representation for the board.

        It uses Unicode code points
        for \N{WHITE CHESS PAWN} and \N{BLACK CHESS PAWN}.
        """

        def add_hline(line: str) -> str:
            line = line + ' '
            for _ in self.COLS:
                line = line + '+---'
            line = line + '+\n'
            return line

        s = '\n '
        for c in self.COLS:
            s += '  ' + c + ' '
        s = add_hline(s + ' \n')

        for r in range(self.NROWS):
            for c in self.COLS:
                s += f' | {MARKS[self[c][r]]}'
            s += ' |\n'
            s = add_hline(s)
        s += f'Board code: {self.get_code()}\n'
        return s

    def __eq__(self, other_object) -> bool:
        if not isinstance(other_object, Board):
            return NotImplemented
        return hash(self) == hash(other_object)

    def __hash__(self) -> int:
        b256 = Base256()
        code = self.get_code()
        return b256.to_int(code)

    def get_code(self) -> Code:
        """Return the code which identifies this board."""
        return pawns_to_code(self.pawns)

    def __getitem__(self, col: ColName) -> list[Player]:
        return self.pawns[col]

    def __setitem__(self, col: ColName, value: list[Player]):
        raise TypeError("'Board' objects are immutable!")

    def flip(self) -> 'Board':
        """Return a board with the columns flipped.
        """
        return Board(self.get_code()[::-1])

    def flip_move(self, move: str):
        """Return a flipped move,
           i.e. the move according to the board with flipped columns.
        """
        r = ''
        for c in move:
            for i, k in enumerate(self.COLS):
                if c == k:
                    r = r + self.COLS[-(i+1)]
        return r

    def exchange(self) -> 'Board':
        """Return a board with players exchanged
           as if VIOLET played as CYAN and viceversa.
        """
        b = {}
        for c in self.COLS:
            b[c] = [EMPTY] * len(self.COLS)
            for i, p in enumerate(self[c][::-1]):
                b[c][i] = other(p)
            b[c] = b[c]
        return Board(pawns_to_code(b))

    def can_move_fwd(self, player: Player, column: ColName) -> bool:
        """True if player can move forward in column.
        """
        if player == VIOLET:
            for row in range(self.NROWS-1):
                if self[column][row] == VIOLET and self[column][row+1] == EMPTY:
                    return True
            return False
        return self.exchange().can_move_fwd(VIOLET, column)

    def can_capture(self, player: Player, column: ColName) -> bool:
        """True if player can move diagonally from column and capture a pawn.
        """
        if player == VIOLET:
            if column == self.COLS[0]:
                for row in range(Board.NROWS-1):
                    if self[self.COLS[0]][row] == VIOLET \
                       and self[self.COLS[1]][row + 1] == CYAN:
                        return True
            elif column == self.COLS[-1]:
                return self.flip().can_capture(VIOLET, self.COLS[0])
            else:
                c = self.COLS.index(column)
                for row in range(self.NROWS-1):
                    if self[column][row] == VIOLET \
                       and CYAN in (self[self.COLS[c-1]][row + 1],
                                    self[self.COLS[c+1]][row + 1]):
                        return True
            return False
        return self.exchange().can_capture(VIOLET, column)

    def is_winner(self, player: Player) -> bool:
        """True if player is winning. """
        if player == VIOLET:
            opponent_moves = 0
            for c in self.COLS:
                if self[c][-1] == VIOLET:
                    return True
                if self.can_move_fwd(CYAN, c) or self.can_capture(CYAN, c):
                    opponent_moves += 1
            return opponent_moves == 0
        return self.exchange().is_winner(VIOLET)

    class IllegalMoveError(Exception):
        """An illegal move was tried."""

    def move(self, player: Player,
             start: ColName,
             end: ColName):
        """Return a board in which player has moved from column start to end.

        Raise an exception if the move is not legal.
        """
        assert start+end in self.WELL_FORMED_MOVES  # pragma: no mutate

        b = {}
        for c in self.COLS:
            b[c] = self[c]

        if player == VIOLET:
            if start == end and not self.can_move_fwd(VIOLET, start):
                raise Board.IllegalMoveError(
                    f"{VIOLET} cannot move forward on column {start}")
            if start == end:
                if not self.can_move_fwd(VIOLET, start):
                    raise Board.IllegalMoveError('Invalid forward move.')
                col = list(self[start])
                i = len(col) - 1 - col[::-1].index(VIOLET)
                col[i] = EMPTY
                col[i + 1] = VIOLET
                b[start] = col
                return Board(pawns_to_code(b))

            if not self.can_capture(VIOLET, start):
                raise Board.IllegalMoveError('Invalid capture move.')

            def capture(bdict: PawnColumn,
                        from_row: int, to_row: int) -> Board | None:
                if self[start][from_row] == VIOLET \
                   and self[end][to_row] == CYAN:
                    bdict[start] = list(self[start])
                    bdict[end] = list(self[end])
                    bdict[start][from_row] = EMPTY
                    bdict[end][to_row] = VIOLET
                    bdict[start] = bdict[start]
                    bdict[end] = bdict[end]
                    return Board(pawns_to_code(bdict))
                return None

            for row in range(self.NROWS-1):
                r = capture(b, row, row + 1)
                if r is not None:
                    return r

            raise Board.IllegalMoveError("No move is possible!")

        return self.exchange().move(VIOLET, start, end).exchange()

    def is_symmetrical(self) -> bool:
        """True if board is symmetric."""
        return self == self.flip()

    def get_legal_moves(self, player: Player) -> list[str]:
        """Return a list of all legal moves for player in a given position."""
        r = []
        for m in self.WELL_FORMED_MOVES:
            try:
                self.move(player, m[0], m[1])
                r.append(m)
            except Board.IllegalMoveError:
                pass
        return r

    def prune_sym_moves(self, moves: list[str]) -> list[str]:
        """Return a list of moves filtered for symmetrical ones."""
        if len(moves) <= 1:
            return moves
        if self.flip_move(moves[0]) in moves[1:]:
            return self.prune_sym_moves(moves[1:])
        return self.prune_sym_moves(moves[1:]) + [moves[0]]

    @classmethod
    def get_forest(cls, ncols: int) -> tuple[str,
                                             dict[str,
                                                  tuple[Player, set[str]]]]:
        """Return a LaTeX string with the game tree
           and the set of unique boards.

        Boards are returned as a dictionary indexed by codes
        with the set of player moves legal from there.
        """

        boards: dict[str, tuple[int, set[str]]] = {}

        def get_tree(board: 'Board', player: Player,
                     turn: Player,
                     a_move: str | None,
                     indent: str) -> tuple[str, Player]:
            code = board.get_code()
            if board.is_winner(other(player)) and a_move is not None:
                s = f"{indent}[{code},winner{other(player)},"
                s += f"move={{{other(player)}}}{{{a_move.lower()}}}]"
                return s, other(player)
            moves = board.get_legal_moves(player)
            if board.is_symmetrical():
                moves = board.prune_sym_moves(moves)
            if code in boards:
                for m in moves:
                    boards[code][1].add(m)
            elif code[::-1] in boards:
                for m in moves:
                    boards[code[::-1]][1].add(board.flip_move(m))
            else:
                boards[code] = (turn, set(moves))
            s = f"{indent}[{code},winner@"
            if a_move is not None:
                s += f",move={{{other(player)}}}{{{a_move.lower()}}},winner@"
            s += "\n"
            winners = []
            for m in moves:
                b = board.move(player, m[0], m[1])
                t, w = get_tree(b, other(player), turn+1, m, indent + " ")
                s += t + "\n"
                winners.append(w)
            if len(set(winners)) == 1:
                winner = winners[0]
            else:
                winner = player
            s = s.replace('@', str(winner))
            return s + indent + "]", winner

        f = """
%% Game tree for LEX --- Learning EX-a-pawn
%% Uncomment the lines which begin with % for a standalone LaTeX document
%\\documentclass[tikz]{standalone}
%\\usepackage{forest}
%\\begin{document}

\\forestset{
  default preamble={
    for tree={font=\\tiny}
  }
}
\\begin{forest}
  winner1/.style={draw,fill=cyan,inner sep=1pt,outer sep=0},
  winner2/.style={draw,fill=violet!60,inner sep=1pt,outer sep=0},
  move/.style n args=2{%
    if={#1<2}%
    {edge label/.expanded={%
      node [midway,fill=white,text=cyan,font=\\unexpanded{\\tiny}] {$#2$}%
    }}{edge label/.expanded={%
      node [midway,fill=white,text=violet!60,font=\\unexpanded{\\tiny}] {$#2$}%
    }},
  }
"""
        t, _ = get_tree(Board(ncols=ncols), CYAN, 1, None, " ")
        f = f + t
        f = f + "\n\\end{forest}"
        f = f + "\n%\\end{document}"

        for h in boards:
            b = Board(h)
            if b.is_symmetrical():
                moves = b.prune_sym_moves(list(boards[h][1]))
                boards[h] = (boards[h][0], set(moves))

        return f, boards



from lexcel import Board, VIOLET, CYAN, EMPTY, pawns_to_code, other, Base256
import unittest


class TestBase256(unittest.TestCase):
    def setUp(self):
        self.b256 = Base256()

    def test_to_int0(self):
        r = self.b256.to_int('0000000')
        self.assertEqual(r, 0)

    def test_to_intBA7(self):
        r = self.b256.to_int('BA7')
        self.assertEqual(r, 7 + 10*256 + 11*256**2)

    def test_to_code1(self):
        r = self.b256.from_int(10)
        self.assertEqual(r, 'A')

    def test_to_code36(self):
        r = self.b256.from_int(10 + 26)
        self.assertEqual(r, 'a')

    def test_to_code255(self):
        r = self.b256.from_int(255)
        self.assertEqual(r, self.b256.SYMBOLS[-1])

    def test_to_code255_255(self):
        r = self.b256.from_int(256**2 - 1)
        self.assertEqual(r, self.b256.SYMBOLS[-1]*2)


class TestLexCommon(unittest.TestCase):
    basic_case = {
        'X': [VIOLET, CYAN, EMPTY],
        'Y': [EMPTY, CYAN, CYAN],
        'Z': [VIOLET, EMPTY, EMPTY]
    }

    def setUp(self):
        self.board = Board(pawns_to_code(self.basic_case))
        self.b256 = Base256()

    def test_str(self):
        r = str(self.board)
        line_len = 4*len(self.board.COLS) + 1
        n_lines = self.board.NROWS * 2 + 2
        caption_len = len("Board code: ") + len(self.board.COLS)
        self.assertEqual(len(r),
                         line_len * n_lines
                         + 2*(n_lines+1)
                         + caption_len)

    def test_to_code_flip(self):
        r = self.board.flip().get_code()
        self.assertEqual(r, self.board.get_code()[::-1])

    def test_items(self):
        self.assertEqual(self.board['X'], self.basic_case['X'])
        with self.assertRaises(TypeError) as e:
            self.board['X'] = None
        self.assertEqual(str(e.exception), "'Board' objects are immutable!")

    def test_exchange(self):
        r = self.board.exchange()
        self.assertEqual(r['X'], [other(i) for i in self.board['X'][::-1]])
        self.assertNotEqual(r, self.board)
        r = r.exchange()
        self.assertEqual(r, self.board)

    def test_flip(self):
        r = self.board.flip()
        self.assertNotEqual(r, self.board)
        for i, c in enumerate(self.board.COLS):
            x = self.board.COLS[i]
            z = self.board.COLS[-(i+1)]
            self.assertEqual(r[x], self.board[z])

    def test_flip_move_prop(self):
        for m in self.board.WELL_FORMED_MOVES:
            fm = self.board.flip_move(m)
            self.assertEqual(m, self.board.flip_move(fm))


class TestLexFunctions(unittest.TestCase):

    def test_pawns_to_code1(self):
        b = {'X': [EMPTY, EMPTY, EMPTY]}
        self.assertEqual(pawns_to_code(b), "0")
        b = {'X': [CYAN, EMPTY, EMPTY]}
        self.assertEqual(pawns_to_code(b), "1")
        b = {'X': [EMPTY, CYAN, EMPTY]}
        self.assertEqual(pawns_to_code(b), "3")
        b = {'X': [EMPTY, EMPTY, CYAN]}
        self.assertEqual(pawns_to_code(b), "9")
        b = {'X': [VIOLET, EMPTY, EMPTY]}
        self.assertEqual(pawns_to_code(b), "2")
        b = {'X': [EMPTY, VIOLET, EMPTY]}
        self.assertEqual(pawns_to_code(b), "6")
        b = {'X': [EMPTY, EMPTY, VIOLET]}
        self.assertEqual(pawns_to_code(b), "I")

    def test_pawns_to_code2(self):
        w = {
            'X': [EMPTY, VIOLET, EMPTY],
            'Y': [EMPTY, EMPTY, VIOLET],
            'Z': [EMPTY, CYAN, EMPTY]
        }
        self.assertEqual(pawns_to_code(w), "6I3")

    def test_other(self):
        self.assertEqual(other(VIOLET), CYAN)
        self.assertEqual(other(CYAN), VIOLET)
        self.assertEqual(other(EMPTY), EMPTY)


class TestBoard3(TestLexCommon):

    def test_hash(self):
        b1 = Board()
        b2 = Board("5C2")
        self.assertEqual(hash(self.board),
                         self.b256.to_int("5C2"))
        self.assertNotEqual(b1, self.board)
        self.assertEqual(b2, self.board)

    def test_can_move_fwd_violet(self):
        self.assertTrue(self.board.can_move_fwd(VIOLET, 'Z'))
        self.assertFalse(self.board.can_move_fwd(VIOLET, 'X'))

    def test_can_move_fwd_cyan(self):
        self.assertTrue(self.board.can_move_fwd(CYAN, 'Y'))
        self.assertFalse(self.board.can_move_fwd(CYAN, 'X'))

    def test_can_capture_violet(self):
        self.assertTrue(self.board.can_capture(VIOLET, 'X'))
        self.assertTrue(self.board.can_capture(VIOLET, 'Z'))
        self.assertFalse(self.board.can_capture(VIOLET, 'Y'))

    def test_can_capture_cyan(self):
        self.assertFalse(self.board.can_capture(CYAN, 'X'))
        self.assertFalse(self.board.can_capture(CYAN, 'Z'))
        self.assertTrue(self.board.can_capture(CYAN, 'Y'))

    def test_winner1(self):
        w = {
            'X': [EMPTY, VIOLET, EMPTY],
            'Y': [EMPTY, EMPTY, VIOLET],
            'Z': [EMPTY, CYAN, EMPTY]
        }
        vw = Board(pawns_to_code(w))
        self.assertTrue(vw.is_winner(VIOLET))
        self.assertFalse(vw.is_winner(CYAN))

    def test_winner2(self):
        w = {
            'X': [EMPTY, VIOLET, CYAN],
            'Y': [EMPTY, EMPTY, EMPTY],
            'Z': [VIOLET, CYAN, EMPTY]
        }
        cvw = Board(pawns_to_code(w))
        for c in w:
            self.assertFalse(cvw.can_move_fwd(CYAN, c))
            self.assertFalse(cvw.can_capture(CYAN, c))

        self.assertTrue(cvw.is_winner(VIOLET))
        self.assertTrue(cvw.is_winner(CYAN))

    def test_move1(self):
        r = self.board.move(VIOLET, 'Z', 'Z')
        self.assertEqual(r['X'], self.board['X'])
        self.assertEqual(r['Y'], self.board['Y'])
        self.assertEqual(r['Z'], [EMPTY, VIOLET, EMPTY])

    def test_move2(self):
        r = self.board.move(VIOLET, 'X', 'Y')
        self.assertEqual(r['X'], [EMPTY, CYAN, EMPTY])
        self.assertEqual(r['Y'], [EMPTY, VIOLET, CYAN])
        self.assertEqual(r['Z'], self.board['Z'])

    def test_move3(self):
        r = self.board.move(VIOLET, 'Z', 'Y')
        self.assertEqual(r['X'], self.board['X'])
        self.assertEqual(r['Y'], [EMPTY, VIOLET, CYAN])
        self.assertEqual(r['Z'], [EMPTY, EMPTY, EMPTY])

    def test_move4(self):
        b = {
            'X': [EMPTY, CYAN, EMPTY],
            'Y': [VIOLET, VIOLET, EMPTY],
            'Z': [VIOLET, EMPTY, CYAN]
        }
        board = Board(pawns_to_code(b))
        r = board.move(VIOLET, 'Y', 'Y')
        self.assertEqual(r['X'], board['X'])
        self.assertEqual(r['Y'], [VIOLET, EMPTY, VIOLET])
        self.assertEqual(r['Z'], board['Z'])

    def test_move5(self):
        b = {
            'X': [EMPTY, CYAN, EMPTY],
            'Y': [VIOLET, VIOLET, EMPTY],
            'Z': [VIOLET, EMPTY, CYAN]
        }
        board = Board(pawns_to_code(b))
        r = board.move(VIOLET, 'Y', 'X')
        self.assertEqual(r['X'], [EMPTY, VIOLET, EMPTY])
        self.assertEqual(r['Y'], [EMPTY, VIOLET, EMPTY])
        self.assertEqual(r['Z'], board['Z'])

    def test_move6(self):
        b = {
            'X': [EMPTY, CYAN, EMPTY],
            'Y': [VIOLET, VIOLET, EMPTY],
            'Z': [VIOLET, EMPTY, CYAN]
        }
        board = Board(pawns_to_code(b))
        r = board.move(VIOLET, 'Y', 'Z')
        self.assertEqual(r['X'], board['X'])
        self.assertEqual(r['Y'], [VIOLET, EMPTY, EMPTY])
        self.assertEqual(r['Z'], [VIOLET, EMPTY, VIOLET])

    def test_move7(self):
        b = {
            'X': [EMPTY, VIOLET, CYAN],
            'Y': [VIOLET, CYAN, EMPTY],
            'Z': [VIOLET, CYAN, EMPTY]
        }
        board = Board(pawns_to_code(b))
        r = board.move(VIOLET, 'Y', 'Z')
        self.assertEqual(r['X'], board['X'])
        self.assertEqual(r['Y'], [EMPTY, CYAN, EMPTY])
        self.assertEqual(r['Z'], [VIOLET, VIOLET, EMPTY])

    def test_move8(self):
        b = {
            'X': [EMPTY, VIOLET, CYAN],
            'Y': [VIOLET, CYAN, EMPTY],
            'Z': [VIOLET, CYAN, EMPTY]
        }
        board = Board(pawns_to_code(b))
        r = board.move(VIOLET, 'Z', 'Y')
        self.assertEqual(r['X'], board['X'])
        self.assertEqual(r['Y'], [VIOLET, VIOLET, EMPTY])
        self.assertEqual(r['Z'], [EMPTY, CYAN, EMPTY])

    def test_move9(self):
        b = Board()
        r = b.move(CYAN, 'X', 'X')
        self.assertEqual(r['X'], [VIOLET, CYAN, EMPTY])
        self.assertEqual(r['Y'], b['Y'])
        self.assertEqual(r['Z'], b['Z'])

    def test_is_symmetrical(self):
        b = Board('B5B')
        self.assertTrue(b.is_symmetrical())

    def test_prune_sym_moves(self):
        self.assertEqual(len(self.board.prune_sym_moves(['XX', 'ZZ'])), 1)

    def test_prune_sym_moves3(self):
        r = self.board.prune_sym_moves(['XX', 'XY', 'ZY'])
        self.assertEqual(len(r), 2)

    def test_get_legal_moves1(self):
        r = self.board.get_legal_moves(CYAN)
        self.assertEqual(len(r), 3)
        self.assertIn('YX', r)
        self.assertIn('YY', r)
        self.assertIn('YZ', r)

    def test_get_legal_moves2(self):
        r = self.board.get_legal_moves(VIOLET)
        self.assertEqual(len(r), 3)
        self.assertIn('XY', r)
        self.assertIn('ZY', r)
        self.assertIn('ZZ', r)

    def test_flip_move(self):
        r = self.board.flip_move('XX')
        self.assertEqual(r, 'ZZ')
        r = self.board.flip_move('XY')
        self.assertEqual(r, 'ZY')
        r = self.board.flip_move('YZ')
        self.assertEqual(r, 'YX')
        r = self.board.flip_move('YY')
        self.assertEqual(r, 'YY')

    def test_get_forest(self):
        _, b = Board.get_forest(3)
        turn2 = dict([(x, b[x]) for x in b if b[x][0] == 2])
        self.assertEqual(len(turn2), 2)
        self.assertIn('B5B', turn2)
        self.assertEqual(len(turn2['B5B']), 2)


class TestBoard4(TestLexCommon):
    basic_case = {
        'W': [EMPTY, VIOLET, CYAN, EMPTY],
        'X': [VIOLET, EMPTY, EMPTY, EMPTY],
        'Y': [EMPTY, CYAN, EMPTY, CYAN],
        'Z': [EMPTY, VIOLET, CYAN, EMPTY]
    }

    def test_hash(self):
        b1 = Board(ncols=4)
        b2 = Board("F2UF")
        self.assertEqual(hash(self.board), self.b256.to_int('F2UF'))
        self.assertNotEqual(b1, self.board)
        self.assertEqual(b2, self.board)

    def test_can_capture_violet(self):
        self.assertTrue(self.board.can_capture(VIOLET, 'X'))
        self.assertFalse(self.board.can_capture(VIOLET, 'Z'))
        self.assertFalse(self.board.can_capture(VIOLET, 'Y'))
        self.assertFalse(self.board.can_capture(VIOLET, 'W'))

    def test_can_capture_cyan(self):
        self.assertFalse(self.board.can_capture(CYAN, 'X'))
        self.assertFalse(self.board.can_capture(CYAN, 'Z'))
        self.assertTrue(self.board.can_capture(CYAN, 'Y'))
        self.assertFalse(self.board.can_capture(CYAN, 'W'))

    def test_winner1(self):
        w = {
            'W': [EMPTY, VIOLET, EMPTY, CYAN],
            'X': [EMPTY, VIOLET, CYAN, EMPTY],
            'Y': [EMPTY, EMPTY, EMPTY, VIOLET],
            'Z': [EMPTY, CYAN, EMPTY, EMPTY]
        }
        vw = Board(pawns_to_code(w))
        self.assertTrue(vw.is_winner(VIOLET))
        self.assertFalse(vw.is_winner(CYAN))

    def test_winner2(self):
        w = {
            'W': [EMPTY, EMPTY, EMPTY, EMPTY],
            'X': [EMPTY, VIOLET, CYAN, EMPTY],
            'Y': [EMPTY, EMPTY, EMPTY, EMPTY],
            'Z': [VIOLET, CYAN, EMPTY, EMPTY]
        }
        cvw = Board(pawns_to_code(w))
        for c in w:
            self.assertFalse(cvw.can_move_fwd(CYAN, c))
            self.assertFalse(cvw.can_capture(CYAN, c))

        self.assertTrue(cvw.is_winner(VIOLET))
        self.assertTrue(cvw.is_winner(CYAN))

    def test_is_symmetrical(self):
        b = Board('B55B')
        self.assertTrue(b.is_symmetrical())

    @unittest.skip("old")
    def test_get_legal_moves1(self):
        r = self.board.get_legal_moves(CYAN)
        self.assertEqual(len(r), 5)
        self.assertIn('WW', r)
        self.assertIn('WX', r)
        self.assertIn('YX', r)
        self.assertIn('YY', r)
        self.assertIn('YZ', r)

    @unittest.skip("old")
    def test_get_legal_moves2(self):
        r = self.board.get_legal_moves(VIOLET)
        self.assertEqual(len(r), 4)
        self.assertIn('XW', r)
        self.assertIn('XY', r)
        self.assertIn('ZY', r)
        self.assertIn('ZZ', r)

    @unittest.skip("old")
    def test_flip_move(self):
        r = self.board.flip_move('WW')
        self.assertEqual(r, 'ZZ')
        r = self.board.flip_move('WX')
        self.assertEqual(r, 'ZY')
        r = self.board.flip_move('XY')
        self.assertEqual(r, 'YX')
        r = self.board.flip_move('YY')
        self.assertEqual(r, 'XX')

@unittest.skip("old")
class TestBoard5(TestLexCommon):
    basic_case = {'V': (VIOLET, EMPTY, CYAN),
                  'W': (EMPTY, CYAN, EMPTY),
                  'X': (VIOLET, CYAN, EMPTY),
                  'Y': (EMPTY, CYAN, CYAN),
                  'Z': (VIOLET, EMPTY, EMPTY)
    }

    def test_hash(self):
        b1 = Board(ncols=5)
        b2 = Board("B35C2")
        #self.assertEqual(hash(self.board), int("B35C2", len(DIGITS)))
        self.assertNotEqual(b1, self.board)
        self.assertEqual(b2, self.board)

    def test_can_capture_violet(self):
        self.assertTrue(self.board.can_capture(VIOLET, 'V'))
        self.assertFalse(self.board.can_capture(VIOLET, 'W'))
        self.assertTrue(self.board.can_capture(VIOLET, 'X'))
        self.assertFalse(self.board.can_capture(VIOLET, 'Y'))
        self.assertTrue(self.board.can_capture(VIOLET, 'Z'))

    def test_can_capture_cyan(self):
        self.assertFalse(self.board.can_capture(CYAN, 'V'))
        self.assertTrue(self.board.can_capture(CYAN, 'W'))
        self.assertFalse(self.board.can_capture(CYAN, 'X'))
        self.assertTrue(self.board.can_capture(CYAN, 'Y'))
        self.assertFalse(self.board.can_capture(CYAN, 'Z'))

    def test_is_symmetrical(self):
        b = Board('B525B')
        self.assertTrue(b.is_symmetrical())

    def test_get_legal_moves1(self):
        r = self.board.get_legal_moves(CYAN)
        self.assertEqual(len(r), 7)
        self.assertIn('VV', r)
        self.assertIn('WV', r)
        self.assertIn('WW', r)
        self.assertIn('WX', r)
        self.assertIn('YX', r)
        self.assertIn('YY', r)
        self.assertIn('YZ', r)

    def test_get_legal_moves2(self):
        r = self.board.get_legal_moves(VIOLET)
        self.assertEqual(len(r), 6)
        self.assertIn('VV', r)
        self.assertIn('VW', r)
        self.assertIn('XW', r)
        self.assertIn('XY', r)
        self.assertIn('ZY', r)
        self.assertIn('ZZ', r)

    def test_flip_move(self):
        r = self.board.flip_move('VV')
        self.assertEqual(r, 'ZZ')
        r = self.board.flip_move('WX')
        self.assertEqual(r, 'YX')
        r = self.board.flip_move('XY')
        self.assertEqual(r, 'XW')
        r = self.board.flip_move('XX')
        self.assertEqual(r, 'XX')


if __name__ == '__main__':
    testSuite = unittest.TestSuite()
    testSuite.addTest(unittest.makeSuite(TestBase256))
    testSuite.addTests(unittest.makeSuite(TestLexFunctions))
    testSuite.addTests(unittest.makeSuite(TestBoard3))
    testSuite.addTests(unittest.makeSuite(TestBoard4))
    testSuite.addTests(unittest.makeSuite(TestBoard5))
    unittest.TextTestRunner().run(testSuite)



"""Flask application for LEX game."""

import os
import random
import logging
import uuid
from flask import Flask, render_template, request, jsonify, session
from werkzeug.utils import secure_filename
from lexcel.lex import Board, CYAN, VIOLET, other
from lexcel.game import get_move, learn

UPLOAD_FOLDER = 'uploads'
ALLOWED_EXTENSIONS = {'xlsx'}

# In-memory storage for game states
# In production, use Redis or a database
games = {}


def allowed_file(filename: str) -> bool:
    """Check if file extension is allowed."""
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS


def get_game_state():
    """Get current game state from server-side storage."""
    game_id = session.get('game_id')
    if game_id and game_id in games:
        return games[game_id]
    return None


def save_game_state(state: dict):
    """Save game state to server-side storage."""
    if 'game_id' not in session:
        session['game_id'] = str(uuid.uuid4())
    games[session['game_id']] = state


def get_player_type(state: dict, player: int) -> str:
    """Get player type (human or computer)."""
    if player == CYAN:
        return state.get('player1_type', 'human')
    return state.get('player2_type', 'human')


def get_player_exp(state: dict, player: int) -> str | None:
    """Get player experience file path."""
    if player == CYAN:
        return state.get('player1_exp')
    return state.get('player2_exp')


def apply_learning(state: dict, winner: int):
    """Apply learning to computer players."""
    # Learn for player 1 (CYAN)
    if state.get('player1_type') == 'computer':
        exp_start = state.get('player1_exp')
        exp_save = state.get('player1_exp')  # Save to same file
        if exp_start and exp_save:
            learn(CYAN, winner,
                  state['memory'][CYAN],
                  exp_start, exp_save)

    # Learn for player 2 (VIOLET)
    if state.get('player2_type') == 'computer':
        exp_start = state.get('player2_exp')
        exp_save = state.get('player2_exp')  # Save to same file
        if exp_start and exp_save:
            learn(VIOLET, winner,
                  state['memory'][VIOLET],
                  exp_start, exp_save)


def board_to_dict(board: Board) -> dict:
    """Convert Board to dictionary for JSON serialization."""
    return {
        'code': board.get_code(),
        'cols': board.COLS,
        'nrows': board.NROWS,
        'pawns': {col: board[col] for col in board.COLS},
        'is_symmetrical': board.is_symmetrical()
    }


def initialize_board(board_code: str | None) -> Board:
    """Initialize a board from code or create a new 3x3 board."""
    if board_code:
        return Board(board_code)
    return Board(ncols=3)


def create_initial_state(board: Board,
                        player1_type: str, player2_type: str,
                        player1_exp: str | None, player2_exp: str | None) -> dict:
    """Create the initial game state dictionary.

    First player is always CYAN.
    """
    return {
        'board_code': board.get_code(),
        'current_player': CYAN,
        'turn': 1,
        'winner': None,
        'memory': {CYAN: [], VIOLET: []},
        'player1_type': player1_type,
        'player2_type': player2_type,
        'player1_exp': player1_exp,
        'player2_exp': player2_exp,
        'random_state': random.Random().getstate()
    }


def handle_computer_move(state: dict, board: Board, current_player: int,
                        turn: int, rand: random.Random, exp_file: str) -> tuple[Board, int | None]:
    """Handle a computer player's move."""
    move_memory = state['memory'][current_player]
    new_board, new_winner = get_move(
        board, current_player, turn, rand,
        exp_file, move_memory
    )
    state['memory'][current_player] = move_memory
    return new_board, new_winner


def handle_human_move(board: Board, current_player: int, move: str) -> tuple[Board, int | None]:
    """Handle a human player's move."""
    move = move.upper()
    if move not in board.WELL_FORMED_MOVES:
        raise ValueError('Invalid move format')

    new_board = board.move(current_player, move[0], move[1])
    new_winner = current_player if new_board.is_winner(current_player) else None
    return new_board, new_winner


def update_game_state_after_move(state: dict, new_board: Board,
                                 new_winner: int | None, turn: int,
                                 rand: random.Random) -> list[str]:
    """Update game state after a move and return next legal moves."""
    state['random_state'] = rand.getstate()
    state['board_code'] = new_board.get_code()
    state['turn'] = turn + 1
    state['winner'] = new_winner

    if not new_winner:
        state['current_player'] = other(state['current_player'])
        return new_board.get_legal_moves(other(state['current_player']))

    apply_learning(state, new_winner)
    return []


def save_uploaded_file(file, player: str, state: dict | None) -> tuple[str, str]:
    """Save uploaded experience file and update game state."""
    filename = secure_filename(file.filename)
    filepath = os.path.join(UPLOAD_FOLDER, filename)
    file.save(filepath)

    if state:
        if player == 'player1':
            state['player1_exp'] = filepath
        else:
            state['player2_exp'] = filepath
        save_game_state(state)

    return filename, filepath


def get_available_experience_files() -> list[str]:
    """Get list of available experience files in upload folder."""
    if not os.path.exists(UPLOAD_FOLDER):
        return []

    files = []
    for filename in os.listdir(UPLOAD_FOLDER):
        if allowed_file(filename):
            files.append(filename)

    return sorted(files)


def create_app(test_config=None) -> Flask:
    """Create and configure the Flask app."""
    app = Flask(__name__)
    app.config['SECRET_KEY'] = os.environ.get('SECRET_KEY', 'dev-secret-key')
    app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
    app.config['MAX_CONTENT_LENGTH'] = 32 * 1024  # 32KB max file size

    if test_config is not None:
        app.config.update(test_config)

    # Ensure upload folder exists
    os.makedirs(UPLOAD_FOLDER, exist_ok=True)

    @app.route('/')
    def index():
        """Render the main game page."""
        return render_template('index.html')

    @app.route('/api/new_game', methods=['POST'])
    def new_game():
        """Start a new game."""
        data = request.get_json()

        board_code = data.get('board_code', None)
        player1_type = data.get('player1_type', 'human')
        player2_type = data.get('player2_type', 'human')
        player1_exp = data.get('player1_exp', None)
        player2_exp = data.get('player2_exp', None)

        board = initialize_board(board_code)
        state = create_initial_state(board, player1_type, player2_type,
                                    player1_exp, player2_exp)
        save_game_state(state)

        return jsonify({
            'success': True,
            'board': board_to_dict(board),
            'current_player': CYAN,
            'turn': 1,
            'legal_moves': board.get_legal_moves(CYAN)
        })

    @app.route('/api/move', methods=['POST'])
    def make_move():
        """Make a move (human or computer)."""
        state = get_game_state()
        if not state:
            return jsonify({'error': 'No game in progress'}), 400

        data = request.get_json()
        move = data.get('move', None)

        board = Board(state['board_code'])
        current_player = state['current_player']
        turn = state['turn']
        winner = state.get('winner')

        if winner:
            return jsonify({'error': 'Game is already finished'}), 400

        player_type = get_player_type(state, current_player)
        rand = random.Random()
        if 'random_state' in state:
            rand.setstate(state['random_state'])

        exp_file = get_player_exp(state, current_player)

        try:
            if player_type == 'computer' and exp_file and move is None:
                new_board, new_winner = handle_computer_move(
                    state, board, current_player, turn, rand, exp_file
                )
            elif move:
                new_board, new_winner = handle_human_move(
                    board, current_player, move
                )
            else:
                return jsonify({'error': 'Move required for human player'}), 400

            next_legal_moves = update_game_state_after_move(
                state, new_board, new_winner, turn, rand
            )
            save_game_state(state)

            return jsonify({
                'success': True,
                'board': board_to_dict(new_board),
                'current_player': state['current_player'],
                'turn': state['turn'],
                'winner': new_winner,
                'legal_moves': next_legal_moves
            })

        except Board.IllegalMoveError as e:
            return jsonify({'error': str(e)}), 400
        except ValueError as e:
            return jsonify({'error': str(e)}), 400
        except Exception as e:
            logging.error("Error making move: %s", e)
            return jsonify({'error': str(e)}), 500

    @app.route('/api/upload_experience', methods=['POST'])
    def upload_experience():
        """Upload an experience Excel file."""
        if 'file' not in request.files:
            return jsonify({'error': 'No file provided'}), 400

        file = request.files['file']
        player = request.form.get('player', 'player1')

        if file.filename == '':
            return jsonify({'error': 'No file selected'}), 400

        if file and allowed_file(file.filename):
            state = get_game_state()
            filename, filepath = save_uploaded_file(file, player, state)

            return jsonify({
                'success': True,
                'filename': filename,
                'filepath': filepath
            })

        return jsonify({'error': 'Invalid file type'}), 400

    @app.route('/api/experience_files', methods=['GET'])
    def list_experience_files():
        """Get list of available experience files."""
        files = get_available_experience_files()
        return jsonify({
            'success': True,
            'files': files
        })

    return app


def main():
    """Run the Flask development server."""
    app = create_app()
    debug = os.environ.get('FLASK_DEBUG', 'true')
    port = os.environ.get('FLASK_PORT', '5000')
    if debug == 'true':
        app.run(debug=True, port=int(port))
    else:
        app.run(debug=False, host='0.0.0.0', port=int(port))

if __name__ == '__main__':
    main()



import argparse
import random
import logging
from lexcel import MARKS, CYAN, VIOLET, other, LExcel, Board, Player


def get_move(board: Board,
             player: Player,
             turn: Player,
             rand: random.Random,
             source: str,
             move_memory: list[tuple[str, str, Player]]):
    b = None
    while b is None:
        if source == 'input':
            logging.debug("From input")  # pragma: no mutate
            m = input(f'Player {player} ({MARKS[player]}), make your move: ')
        else:
            logging.debug("Choose %s %s %s %s %s",
                         source, board.get_code(), turn,
                         move_memory, player)  # pragma: no mutate
            logging.debug("Random %s", rand.getstate())  # pragma: no mutate
            m = LExcel(source).choose_and_remember_move(board,
                                                        turn,
                                                        move_memory,
                                                        rand,
                                                        player)
        m = m.upper()
        logging.debug("Move: %s", m)  # pragma: no mutate
        if m in board.WELL_FORMED_MOVES:
            try:
                b = board.move(player, m[0], m[1])
            except Board.IllegalMoveError:
                print(f'{m} move for player {player} is not allowed.')
                print(board)
        else:
            print('Input is invalid.')

    print(b)
    if b.is_winner(player):
        print(f'Player {player} won!!!')
        return b, player
    return b, None


def learn(a_player: Player,
          the_winner: Player,
          mem: list[tuple[str, str, Player]],
          start: str, to: str | None):
    if to is not None:
        start_lex = LExcel(start)
        if the_winner == a_player:
            # reward each move in memory
            start_lex.reward(to, mem, 1)
        else:
            assert the_winner == other(a_player)  # pragma: no mutate
            # punish last move in memory
            start_lex.reward(to, mem[-1:], -1)


def main(test_args=None):
    desc = """
LEX --- Learning EX-a-pawn
A game written by Mattia Monga for a 'Coding for lawyers' course.
© 2020-25 - Free to distribute, use, and modify
according to the terms of GPLv3."""

    rules = """Pawns move and capture as in chess,
but there are neither two-step moves nor en-passant captures.
Players win by reaching the last row or by blocking the opponent.
Moves are given by two letters: the starting column and the ending one.
"""

    parser = argparse.ArgumentParser(description=desc+'\n\n'+rules,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-e1', '--exp-player1', nargs=2,
                        metavar=('start.xlsx', 'save.xlsx'),
                        default=['input', None],
                        help='Spreadsheet files to get and save game experience used by player 1')
    parser.add_argument('-e2', '--exp-player2', nargs=2,
                        metavar=('start.xlsx', 'save.xlsx'),
                        default=['input', None],
                        help='Spreadsheet files to get and save game experience used by player 2')

    parser.add_argument('-s', '--seed', type=int,
                        help="""Set a seed for random number generations
                        (only useful together with -e1 or -e2)""")

    dim_args = parser.add_mutually_exclusive_group()
    dim_args.add_argument('-b', '--board', type=str,
                          help='Set the initial board by giving its code (default "BBB")')
    dim_args.add_argument('-n', '--num', type=int,
                          default=3, choices=range(3, 4),
                          help='Set the number of pawns each player controls (default 3)')

    parser.add_argument('-p', '--player', type=int,
                        default=CYAN, choices=(CYAN, VIOLET),
                        help='Set the first player (default CYAN)')
    parser.add_argument('-t', '--turn', type=int,
                        default=1, choices=range(1, 11),
                        help='Set the turn (default 1)')

    only_args = parser.add_mutually_exclusive_group()
    only_args.add_argument('--tree', action='store_true',
                           help='Print the game tree in LaTeX')

    only_args.add_argument('--empty', type=str, metavar='empty.xlsx',
                           help='Save a spreadsheet with an empty experience')

    parser.add_argument('-d', '--debug', type=str, metavar='move.log',
                        help='Log moves in file')

    args = parser.parse_args(test_args)
    if args.debug:
        logging.basicConfig(filename=args.debug,
                            level=logging.DEBUG)  # pragma: no mutate
        logging.debug("CLI Args: %s", args)  # pragma: no mutate

    if args.board is not None:
        board = Board(args.board)
        args.num = len(board.get_code())
    else:
        board = Board(ncols=args.num)
    if args.tree:
        print(Board.get_forest(len(board.COLS))[0])
        exit(0)
    if args.empty:
        LExcel.make_empty_experience(len(board.COLS),
                                     args.empty,
                                     [VIOLET, CYAN])
        exit(0)

    turn, winner = args.turn, None
    memory = {CYAN: [], VIOLET: []}
    randomizer = random.Random(args.seed)

    print(desc)
    print(board)
    print(rules+f"\nWell formed moves: {board.WELL_FORMED_MOVES}")
    while winner is None:
        board, winner = get_move(board, args.player, turn, randomizer,
                                 args.exp_player1[0], memory[args.player])
        turn += 1
        if winner is None:
            board, winner = get_move(board, other(args.player),
                                     turn, randomizer,
                                     args.exp_player2[0],
                                     memory[other(args.player)])
            turn += 1

    learn(CYAN, winner, memory[CYAN],
          args.exp_player1[0], args.exp_player1[1])
    learn(VIOLET, winner, memory[VIOLET],
          args.exp_player2[0], args.exp_player2[1])

    exit(0)


if __name__ == '__main__':
    main()



"""Web interface for LEX board game."""

# Non importiamo nulla qui per evitare il warning
# Se necessario, l'app può essere importata esplicitamente con:
# from lexweb.app import create_app

__all__ = []



import unittest
import os
import random
from lexcel import LExcel, Board, VIOLET, CYAN
from openpyxl.utils import column_index_from_string


class TestLexcel(unittest.TestCase):

    def setUp(self):
        self.EMPTY_EXP = 'tests/exapawn-empty.xlsx'
        self.TMP = 'tests/test-exapawn-tmp.xlsx'
        self.RAND = random.Random(42)
        self.e = LExcel(self.EMPTY_EXP)

    def tearDown(self):
        if os.path.isfile(self.TMP):
            os.remove(self.TMP)

    def test_find_board_column(self):
        cell = self.e.find_board_column(2, '5BB')
        self.assertEqual(cell, column_index_from_string('A'))

        cell = self.e.find_board_column(4, '38B')
        self.assertEqual(cell, column_index_from_string('E'))

        cell = self.e.find_board_column(6, '830')
        self.assertEqual(cell, column_index_from_string('Y'))

    def test_get_moves(self):
        moves, weights = self.e.get_moves(Board('38B'), 4, VIOLET)
        self.assertEqual(len(moves), 6)
        self.assertIn('YY', moves)
        self.assertIn('YX', moves)
        self.assertIn('YZ', moves)
        self.assertIn('ZZ', moves)
        self.assertIn('XX', moves)  # illegal, weight 0
        self.assertIn('XY', moves)  # illegal, weight 0
        self.assertEqual(sum(weights), 4)

    def test_choose_and_remember_move(self):
        match = []
        move = self.e.choose_and_remember_move(Board('895'),
                                               4, match, self.RAND, VIOLET)
        self.assertTrue(move == 'XX' or move == 'XY')
        self.assertEqual(match, [(move, '895', 4)])

    def test_choose_and_remember_move_reproducibile(self):
        match = []
        myrand = random.Random(666)
        old = []
        for j in range(10):
            old.append(self.e.choose_and_remember_move(Board('83B'),
                                                       4,
                                                       match, myrand, VIOLET))

        for i in range(2):
            myrand = random.Random(666)
            for j in range(10):
                move = self.e.choose_and_remember_move(Board('83B'),
                                                       4,
                                                       match, myrand, VIOLET)
                self.assertEqual(move, old[j])

    def test_reward(self):
        match = []
        self.e.choose_and_remember_move(Board('895'),
                                        4, match, self.RAND, VIOLET)
        self.e.reward(self.TMP, match, 1)
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board('895'), 4, VIOLET)
        self.assertIn(2, weights)

    def test_punish(self):
        match = []
        self.e.choose_and_remember_move(Board('895'),
                                        4, match, self.RAND, VIOLET)
        self.e.reward(self.TMP, match, -1)
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board('895'), 4, VIOLET)
        self.assertIn(0, weights)

    def test_get_moves_flipped(self):
        r = self.e.get_moves(Board('B83'), 4, VIOLET)
        self.assertIsNone(r)

    def test_choose_and_remember_move_flipped(self):
        match = []
        move = self.e.choose_and_remember_move(Board('335'),
                                               6, match, self.RAND, VIOLET)
        self.assertEqual(move, 'ZY')
        self.assertEqual(match, [(Board('533').flip_move(move), '533', 6)])
        col = self.e.find_board_column(6, match[0][1])
        self.assertIsNotNone(col)

    def test_reward_flipped(self):
        match = []
        self.e.choose_and_remember_move(Board('55F'), 4, match, self.RAND, VIOLET)
        self.e.reward(self.TMP, match, 1)
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board('F55'), 4, VIOLET)
        self.assertIn(2, weights)

    def test_make_empty_experience(self):
        LExcel.make_empty_experience(len('B38'), self.TMP, [VIOLET, CYAN])
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board('B38'), 4, VIOLET)
        self.assertEqual(len(moves), 4)
        self.assertEqual(sum(weights), 4)

    @unittest.skip("old")
    def test_make_empty_experience4(self):
        LExcel.make_empty_experience(4, self.TMP, [VIOLET, CYAN])
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board(ncols=4), 1, CYAN)
        self.assertEqual(len(moves), 2)
        self.assertEqual(sum(weights), 2)

    @unittest.skip("old")
    def test_make_empty_experience5(self):
        LExcel.make_empty_experience(5, self.TMP, [VIOLET, CYAN])
        self.assertTrue(os.path.isfile(self.TMP))
        t = LExcel(self.TMP)
        moves, weights = t.get_moves(Board(ncols=5), 1, CYAN)
        self.assertEqual(len(moves), 3)
        self.assertEqual(sum(weights), 3)


if __name__ == '__main__':
    unittest.main()
